/*=======================================================================*
 * A useless (?) program to convert a QL screen, not yet QXL compatible, *
 * from address $20000 (131072) to a PC Windows 3.1 Bitmap file (BMP).   *
 * The resulting file can be loaded into Windows Paintbrush and played   *
 * with.                                                                 *
 *-----------------------------------------------------------------------*
 * Copyright Norman Dunbar June 1994.                                    *
 *-----------------------------------------------------------------------*
 * BUGS OUTSTANDING.                                                     *
 *                                                                       *
 * 1. Not my fault ! When saving the resulting file in PAINTBRUSH or any *
 *    other bitmap manipulation program, it causes a general protection  *
 *    exception on the PC. This is a major bummer, BUT, only if saved as *
 *    a bitmap file. It saves perfectly well in any other format, eg PCX *
 *    and I don't have a bloody clue as to why it causes a problem as a  *
 *    bmp file. I do note that I create a 16 colour bitmap, but windows  *
 *    tries to save it as a 256 colour bmp (or pcx) which really expands *
 *    the output file. So much for data compression !                    *
 *                                                                       *
 *    BUGFIX ! - load into Paintbrush, CUT whole thing to clipboard,     *
 *    FILE | NEW | PASTE | SAVE_AS samename works ok and no errors.      *
 *-----------------------------------------------------------------------*
 * HISTORY.                                                              *
 *                                                                       *
 * 1. First attempt seemed to work ok but colours all wrong on the PC.   *
 * 2. Second attempt, colours ok now, it was a case of not writing to    *
 *    the disc file properly.                                      v1.01 *
 * 3. 23/06/94 - added mode 8 handling.                            v1.02 *
 * 4. 24/06/94 - Doubled pixel width in mode 8 to keep proportions.v1.03 *
 * 5. 22/08/94 - Added command line parameters. Output adds _BMP.  v1.04 *
 *=======================================================================*/

#include <stdio_h>
#include <qdos_h>
#include <string_h>
#include <stdlib_h>


long (*_cmdchannels)() = NULL;                 /* No channels redirected */
long (*_stackchannels)() = NULL;               /* Or on stack either     */
int  (*_conwrite)() = NULL;                    /* Or output translation  */
void (*_cmdwildcard)() = cmdexpand;            /* Allow wildcard params  */

void consetup_title();                         /* Fancy title window */
void (*_consetup)() = consetup_title;          /* Link it in         */

char _prog_name[] = "Scr2bmp";                 /* Program name       */
char _version[] = "Version 1.04";              /* Version number     */
char _copyright[] = " Norman Dunbar, 1994";   /* Author & copright  */

void write_BMfile_header(FILE *fp);            /* Write BITMAPFILEHEADER */
void write_BMInfo(FILE *fp);                   /* Write BITMAPIFO        */
void write_RGBQuad(FILE *fp);                  /* Write pallette info    */
void load_screen(FILE *fp);                    /* Load QL screen file    */
void convert_screen(FILE *fp, int mode);       /* Convert it to _BMP     */
void set_mode(int mode);                       /* Check & set QL mode    */


/*=====================================================================MAIN==*/
main(int argc, char *argv[])
{
   /*-------------------------------------------*
    * The program starts here for non C readers *
    *-------------------------------------------*/

   char bmp_file[128];     /* Buffer for QL screen filename   */
   char scr_file[128];     /* Buffer for PC BMP filename      */
   int  mode;              /* Holds mode 4 or 8 accordingly   */
   FILE *scr = NULL;       /* File pointer for QL screen file */
   FILE *bmp = NULL;       /* File pointer for PC BMP file    */
   int  file;              /* Loop counter for each parameter */

   /*----------------------------------------*
    * Did I get at least 3 (+1) parameters ? *
    *----------------------------------------*/

   if (argc < 4) {
       printf("USAGE : ex SCR2BMP;'scr_files mode bmp_device'\n\n");
       printf("Where :\n\n");
       printf("scr_files  = All filenames to be converted to bmp files\n");
       printf("             EG. flp1_*_SCR = all '_scr' files on 'flp1_'\n\n");
       printf("mode       = Screen mode of ALL the supplied filenames\n");
       printf("             EG. 4 or 8 for mode 4 or mode 8\n\n");
       printf("bmp_device = Device where all converted files are written\n");
       printf("             EG. ram1_ = write converted files to 'ram1_'\n\n");
       exit(0);
   }

   /*--------------------------*
    * Set the mode if required *
    *--------------------------*/

   mode = atoi(argv[argc -2]);
   if ((mode != 4) && (mode != 8)) {
       printf("Incorrect mode supplied, must be 4 or 8 only\n");
       exit (0);
   }

   set_mode(mode);

   /*---------------------------------------*
    * A loop to process all the input files *
    *---------------------------------------*/

   for (file = 1; file < argc -2; file++) {
       strcpy(scr_file, argv[file]);

       scr = fopen(scr_file, "r");
       if (scr == NULL) {
           sd_clear(fgetchid(stdout), -1);
           printf("%s can't be opened.\n", scr_file);
           continue;
       }

       strcpy(bmp_file, argv[argc-1]);
       strncat(bmp_file, scr_file +5, strlen(scr_file) -5);
       strcat(bmp_file, "_bmp");

       bmp = fopen(bmp_file, "wb");
       if (bmp == NULL) {
           sd_clear(fgetchid(stdout), -1);
           printf("%s can't be opened.\n", bmp_file);
           fclose(scr);
           continue;
       }

       /*-------------------------------------------------------*
        * Write out the various PC header stuff for a BMP file. *
        *-------------------------------------------------------*/

       write_BMfile_header(bmp);
       write_BMInfo(bmp);
       write_RGBQuad(bmp);

       /*----------------------------------------------------------------*
        * Convert the screen to a BMP format and then close the PC file. *
        *----------------------------------------------------------------*/

       load_screen(scr);
       convert_screen(bmp, mode);

       fclose(bmp);
       fclose(scr);

   } /* end for */

   sd_clear(fgetchid(stdout), -1);
   exit(0);
}



/*=================================================================SET_MODE==*/
void set_mode(int mode)
{
   /*------------------------------------------------*
    * Read the current mode for the QL, if it is     *
    * different from that required for the screen    *
    * that is to be converted, set the correct mode. *
    *------------------------------------------------*/

   short s_mode,   /* Current mode */
         d_type;   /* Display type */


   /*------------------------------*
    * Read the current mode first. *
    *------------------------------*/

   d_type = -1;
   s_mode = -1;

   mt_dmode(&s_mode, &d_type);

   /*-------------------------------------------*
    * Is the current mode the same as we need ? *
    * if not, set it.                           *
    *-------------------------------------------*/

   if (!s_mode) s_mode = 4;    /* Zero = mode 4 */

   if (s_mode != (short)mode) {
       s_mode = mode;
       d_type = -1;
       mt_dmode(&s_mode, &d_type);
   }
}


/*==============================================================LOAD_SCREEN==*/
void load_screen(FILE *fp)
{
   /*-----------------------------------------*
    * Simply read 256 scan lines of 128 bytes *
    * into the QL display memory at 131072.   *
    *-----------------------------------------*/

   fread((char *)131072, 256, 128, fp);
}


/*======================================================WRITE_BMFILE_HEADER==*/
void write_BMfile_header(FILE *fp)
{
   int x;
   static char bmh[] = {  'B', 'M',       /* Flag for Bit Map                */
                          118, 0, 1, 0,   /* File size in bytes (not DWORDS) */
                          0, 0, 0, 0,     /* Reserved 1 & reserved 2         */
                          118, 0, 0, 0    /* Offset (bytes) to the bitmap    */
                       };

   /*------------------------------------------*
    * Simply write the above characters to the *
    * PC bitmap file.                          *
    *------------------------------------------*/

   for (x = 0; x < 14; x++) {
       putc(bmh[x], fp);
   }
}


/*=============================================================WRITE_BMINFO==*/
void write_BMInfo(FILE *fp)
{
   int x;
   static char bmi[] = {  40, 0, 0, 0,    /* Bytes in this header   */
                          0, 2, 0, 0,     /* Width in pixels = 512  */
                          0, 1, 0, 0,     /* Height in pixels = 256 */
                          1, 0,           /* Planes = 1             */
                          4, 0,           /* Bits per pixel = 4     */
                          0, 0, 0, 0,     /* Compression = none     */
                          0, 0, 1, 0,     /* Image size in Bytes    */
                          0, 0, 0, 0,     /* Pixels per metre wide  */
                          0, 0, 0, 0,     /* Pixels per meter high  */
                          0, 0, 0, 0,     /* Colours in index used  */
                          16, 0, 0, 0     /* Important colours used */
                       };

   /*--------------------------------------*
    * Write the above info to the PC file. *
    *--------------------------------------*/

   for (x = 0; x < bmi[0]; x++) {
       putc(bmi[x], fp);
   }
}


/*============================================================WRITE_RGBQUAD==*/
void write_RGBQuad(FILE *fp)
{
   int x;
   static char bmq[] = {  0, 0, 0, 0,       /*  0 = Black      */
                          0, 0, 191, 0,     /*  1 = Dk Red     */
                          0, 191, 0, 0,     /*  2 = Dk Green   */
                          0, 191, 191, 0,   /*  3 = Dk Yellow  */
                          191, 0, 0, 0,     /*  4 = Dk Blue    */
                          191, 0, 191, 0,   /*  5 = Dk Magenta */
                          191, 191, 0, 0,   /*  6 = Dk Cyan    */
                          192, 192, 192, 0, /*  7 = Lt Grey    */
                          128, 128, 128, 0, /*  8 = Dk Grey    */
                          0, 0, 255, 0,     /*  9 = Red        */
                          0, 255, 0, 0,     /* 10 = Green      */
                          0, 255, 255, 0,   /* 11 = Yellow     */
                          255, 0, 0, 0,     /* 12 = Blue       */
                          255, 0, 255, 0,   /* 13 = Magenta    */
                          255, 255, 0, 0,   /* 14 = Cyan       */
                          255, 255, 255 ,0  /* 15 = White      */
                       };

   /*------------------------------*
    * Write out the above colours. *
    *------------------------------*/

   for (x = 0; x < 64; x++) {
       putc(bmq[x], fp);
   }
}


/*===========================================================CONVERT_SCREEN==*/
void convert_screen(FILE *fp, int mode)
{
   int scan_line,          /* Loop counter for each scan line      */
       line_offset,        /* Offset into display memeory for line */
       word,               /* Loop counter for each word in a line */
       word_offset;        /* Offset into memeory for each word    */

   unsigned char   red,    /* Odd addresses byte of word offset    */
                   green,  /* Even addressed byte of word offset   */

                   top4_nibble[] = {0x00,  /* Black */
                                    0x90,  /* Red   */
                                    0xA0,  /* Green */
                                    0xF0}, /* White */

                   bot4_nibble[] = {0x00,  /* Black */
                                    0x09,  /* Red   */
                                    0x0A,  /* Green */
                                    0x0F}, /* White */

                   top8_nibble[] = {0x00,  /* Black   */
                                    0xC0,  /* Blue    */
                                    0x90,  /* Red     */
                                    0xD0,  /* Magenta */
                                    0xA0,  /* Green   */
                                    0xE0,  /* Cyan    */
                                    0xB0,  /* Yellow  */
                                    0xF0}, /* White   */

                   bot8_nibble[] = {0x00,  /* Black   */
                                    0x0C,  /* Blue    */
                                    0x09,  /* Red     */
                                    0x0D,  /* Magenta */
                                    0x0A,  /* Green   */
                                    0x0E,  /* Cyan    */
                                    0x0B,  /* Yellow  */
                                    0x0F}, /* White   */

                   pixel[8];   /* One words worth of pixels */


   /*------------------------------------------*
    * There are 256 scan_lines in a QL screen  *
    * so calculate its position in memory. A   *
    * PC bitmap file scans from lower left up. *
    *------------------------------------------*/

   for (scan_line = 255; scan_line > -1; scan_line--) {
       line_offset = 131072 + (scan_line * 128);

       /*------------------------------------------*
        * There are 64 words in a scan_line. The   *
        * memeory address is the scan_line address *
        * plus the word * 2.                       *
        *------------------------------------------*/

       for (word = 0; word < 64; word++) {
           word_offset = line_offset + (word << 1);

           /*------------------------------------------------------*
            * Fetch the green byte and the red byte for each word. *
            *------------------------------------------------------*/

           green = *((unsigned char *)word_offset);
           red   = *((unsigned char *)(word_offset +1));

           /*------------------------------------------*
            * invert the display to show where we are. *
            *------------------------------------------*/

           *((unsigned char *)word_offset) = mode == 4? ~green: (~green)& 0xAA;
           *((unsigned char *)(word_offset +1)) = ~red;

           /*========== MODE 4 ==========*=========== MODE 8 ===========*
            * GREEN | RED | PIXEL COLOUR * GR | RD | BL | PIXEL COLOUR  *
            *============================*==============================*
            *   0   |  0  | 00 = BLACK   *  0 |  0 |  0 | 000 = BLACK   *
            *   0   |  1  | 01 = RED     *  0 |  0 |  1 | 001 = BLUE    *
            *   1   |  0  | 10 = GREEN   *  0 |  1 |  0 | 010 = RED     *
            *   1   |  1  | 11 = WHITE   *  0 |  1 |  1 | 011 = MAGENTA *
            *----------------------------*  1 |  0 |  0 | 100 = GREEN   *
                                         *  1 |  0 |  1 | 101 = CYAN    *
                                         *  1 |  1 |  0 | 110 = YELLOW  *
                                         *  1 |  1 |  1 | 111 = WHITE   *
                                         *------------------------------*/

           /*--------------------------------------------------*
            * Using the above table, mask each bit out of the  *
            * red and green bytes, shift them along to the far *
            * right but leave the green bit in bit 1 while the *
            * red bit is in bit 0, now OR them together to get *
            * a value between 0 and 3 as above.                *
            *--------------------------------------------------*/

           if (mode == 4) {

               pixel[0] = ((green & 128) >> 6) | ((red & 128) >> 7);
               pixel[1] = ((green &  64) >> 5) | ((red &  64) >> 6);
               pixel[2] = ((green &  32) >> 4) | ((red &  32) >> 5);
               pixel[3] = ((green &  16) >> 3) | ((red &  16) >> 4);
               pixel[4] = ((green &   8) >> 2) | ((red &   8) >> 3);
               pixel[5] = ((green &   4) >> 1) | ((red &   4) >> 2);
               pixel[6] = ((green &   2))      | ((red &   2) >> 1);
               pixel[7] = ((green &   1) << 1) |  (red & 1);

               /*----------------------------------------------------*
                * In a PC BMP file, (uncompressed) each pixel uses 4 *
                * bits. The 4 bits make up the index into the colour *
                * pallette above. The array PIXEL holds data for 8   *
                * pixels so we can write out 4 bytes to the PC file. *
                *----------------------------------------------------*/

               fprintf(fp, "%c%c%c%c",
                            (top4_nibble[pixel[0]] | bot4_nibble[pixel[1]]),
                            (top4_nibble[pixel[2]] | bot4_nibble[pixel[3]]),
                            (top4_nibble[pixel[4]] | bot4_nibble[pixel[5]]),
                            (top4_nibble[pixel[6]] | bot4_nibble[pixel[7]]) );

           } else {

               pixel[0] = ((green & 128) >> 5) | ((red & 192) >> 6);
               pixel[1] = ((green &  32) >> 3) | ((red &  48) >> 4);
               pixel[2] = ((green &   8) >> 1) | ((red &  12) >> 2);
               pixel[3] = ((green &   2) << 1) | ((red &   3));

               fprintf(fp, "%c%c%c%c",
                            (top8_nibble[pixel[0]] | bot8_nibble[pixel[0]]),
                            (top8_nibble[pixel[1]] | bot8_nibble[pixel[1]]),
                            (top8_nibble[pixel[2]] | bot8_nibble[pixel[2]]),
                            (top8_nibble[pixel[3]] | bot8_nibble[pixel[3]]) );
           } /* end if    */
       } /* end for word  */
   } /* end for scan_line */

}
